/*
 * MakeDocJ.java
 *
 * Copyright 2003 by BRiSK Software,
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: MakeDocJ.java,v 1.11 2006/02/10 04:27:28 brisk Exp $
 */

import java.awt.Toolkit;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;

import java.io.*;

import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;
import java.util.Date;

// NOTE: for JDK 1.1.x get special collections
// from Sun http://java.sun.com/products/javabeans/infobus/collectionsreadme.html
// import com.sun.java.util.collections.*;
// NOTE: for JDK 1.2+
// import java.util.*;
import @COLLECTIONS@; // used by ant

import gnu.regexp.*;

/**
 * <code>MakeDocJ</code> is the main application used to manipulate PilotDocs
 * @author Jeffrey A. Krzysztow
 * @author Reinhold F. Kainhofer
 * @author Pat Beirne
 * @version 3.11.1
 */
public final class MakeDocJ {

	final static String version = "3.14.0";

	/**
	 * Decodes the PilotDoc file
	 * @param options DecodeOptions
	 * @throws IOException to the caller can handle any I/O errors
	 * @since 1.0
	 */
	public static boolean decode(DecodeOptions options) throws IOException {
		Date startTime = new Date();
		RandomAccessFile fin = null;
		FileOutputStream fout = null;
		try {
			fin = new RandomAccessFile(options.sourcePath, "r");
			fout = new FileOutputStream(options.destinationFile);
		}
		catch(FileNotFoundException e) {
			System.err.println("The source file " + options.sourcePath + " could not be found");
			return(false);
		}
		catch(IOException e) {
			System.err.println("The destination file " + options.destinationFile + " could not be created");
			return(false);
		}

		DatabaseHeader dbHeader = new DatabaseHeader();
		dbHeader.read(fin);
		if(options.quiet == false) {
			System.out.println(dbHeader);
		}

		if(dbHeader.typeID != DatabaseHeader.TEXt
		   || DatabaseHeader.isKnownCreatorID(dbHeader.creatorID) == false) {
			System.err.println("file is unknown format");
			System.err.println(dbHeader);
			return(false);
		}

		RecordIndex[] ri = new RecordIndex[dbHeader.numRecords];
		for(int i = 0; i < dbHeader.numRecords; i++) {
			ri[i] = new RecordIndex();
			ri[i].read(fin);
		}

		DocumentHeader docHeader = new DocumentHeader();
		fin.seek(ri[0].fileOffset);
		docHeader.read(fin);
		if(options.quiet == false) {
			System.out.println(docHeader);
		}

		if(docHeader.version != DocumentHeader.UNCOMPRESSED && docHeader.version != DocumentHeader.COMPRESSED) {
			System.err.println("WARNING: unknown file compression type: " + docHeader.version);
		}

		boolean compressed = (docHeader.version == DocumentHeader.COMPRESSED);

		/////////////////////////////////////////////////////////////
		// Handle the text (compressed or not)
		PilotDocRecord textBuffer = null;
		int dwRecLen;
		for(int i = 1; i <= docHeader.numRecords; i++) {
			System.out.println("Converting `" + dbHeader.name + "`: text record " + i + " of " + docHeader.numRecords);
			if(i == ri.length - 1) {
				// for the last, use the file len
				dwRecLen = (int) fin.length() - ri[i].fileOffset;
			}
			else {
				dwRecLen = ri[i + 1].fileOffset - ri[i].fileOffset;
			}
			if(textBuffer == null || dwRecLen != textBuffer.length()) {
				textBuffer = new PilotDocRecord(dwRecLen);
			}
			fin.seek(ri[i].fileOffset);
			fin.read(textBuffer.buf);
			if(compressed) {
				textBuffer.decompress();
			}
			if(options.eolType != PilotDocRecord.EOLLF) {
				textBuffer.convertEOL(options.eolType);
			}

			fout.write(textBuffer.buf);
		}

		/////////////////////////////////////////////////////////////
		// Handle Bookmarks if any
		if(options.quiet == false) {
			System.out.println(">> Bookmarks <<");
			System.out.println("Has " + (dbHeader.numRecords - (docHeader.numRecords + 1)) + " bookmarks");
		}
		if(options.bookmarks && docHeader.numRecords + 1 < dbHeader.numRecords) {
			PrintWriter bout;
			try {
				bout = new PrintWriter(new FileOutputStream(options.destinationFile + ".bm"));
			}
			catch(IOException e) {
				System.err.println("The file " + options.destinationFile + ".bm" + " could not be created");
				return(false);
			}
			BookMark bm = new BookMark();
			for(int i = docHeader.numRecords + 1; i < dbHeader.numRecords; i++) {
				fin.seek(ri[i].fileOffset);
				bm.read(fin);
				bout.println(bm.name + "," + bm.fileOffset);
			}
			bout.close();
		}
		fin.close();
		fout.close();
		Date stopTime = new Date();
		System.out.println("Decode time = " + (stopTime.getTime() - startTime.getTime()) + "ms");
		return(true);
	}

	final static char progress[] =  { '-', '\\', '|', '/'};

	/**
	 * Encodes a PilotDoc file
	 * @param options EncodesOptions
	 * @throws IOException to the caller can handle any I/O errors
	 * @since 1.0
	 */
	public static boolean encode(EncodeOptions options) throws IOException {
		Date startTime = new Date();
		boolean deletePilotDoc = false;
		ArrayList bookMarks = new ArrayList();
		PilotDocRecord textBuffer = null;
		RandomAccessFile fout = null;
		DocumentHeader docHeader = new DocumentHeader();
		byte attribute = 0;
		if(options.makePrivate) {
			attribute |= RecordIndex.PRIVATE;
		}
		if(options.title.equals("")) {
			System.err.println("Must specify a valid Title (Palm Database Name)");
			return(false);
		}

		if(options.sourcePath.equalsIgnoreCase("clip:")) {
			Transferable transferable = Toolkit.getDefaultToolkit().getSystemClipboard().getContents(null);

			try {
				if(transferable != null && transferable.isDataFlavorSupported(DataFlavor.stringFlavor)) {
					final String clipboard = (String)transferable.getTransferData(DataFlavor.stringFlavor);
					textBuffer = new PilotDocRecord(0);
					textBuffer.buf = clipboard.getBytes();
				}
			}
			catch(UnsupportedFlavorException e) {
			}
			catch(IOException e) {
			}
		}
		else {
			try {
				FileInputStream fin = new FileInputStream(options.sourcePath);
				docHeader.storyLen = fin.available();
				textBuffer = new PilotDocRecord(docHeader.storyLen);	// allocate enough to hold entire file
				fin.read(textBuffer.buf);	// read entire file
				fin.close();
			}
			catch(IOException e) {
				System.err.println("The source file `" + options.sourcePath + "` could not be found.");
				return(false);
			}
		}

		try {
			options.destinationFile.delete();
			fout = new RandomAccessFile(options.destinationFile, "rw");
		}
		catch(IOException e) {
			System.err.println("The destination file `" + options.destinationFile + "` could not be created.");
			return(false);
		}

		System.out.println("Encoding `" + options.title + "`," +
						   (options.binary ? " binary," : "")  +
						   (options.compress ? "" : " not") + " compressed" +
						   (options.makePrivate ? " private" : "") +
						   (options.unwrap ? " unwrapped" : "")
						  );

		if(options.binary == false) {
			textBuffer.removeBinary();
		}
		docHeader.storyLen = textBuffer.convertEOL();
		if(options.unwrap) {
			docHeader.storyLen = textBuffer.unwrap(options.lineLength);
		}
		if(options.compressEOL) {
			docHeader.storyLen = textBuffer.compressEOL();
		}

		byte[] uncompressedBuffer = new byte[textBuffer.length()];
		System.arraycopy(textBuffer.buf, 0, uncompressedBuffer, 0, textBuffer.length());
		String uncompressedText = new String(uncompressedBuffer);

		///////////////////////////////////////////////////
		// deal with bookmarks from the command line
		if(options.bookMarks.size() > 0) {
			Iterator bookMark = options.bookMarks.listIterator();
			try {
				while(true) {
					processBookMark(uncompressedText, (String)bookMark.next(), bookMarks, options);
				}
			}
			catch(NoSuchElementException e) {
				// we are at the end of the list!
			}
		}

		///////////////////////////////////////////////////
		// deal with bookmark file if necessary
		if(options.bookMarkPath != null) {
			LineNumberReader finb = null;
			try {
				finb = new LineNumberReader(new FileReader(options.bookMarkPath));
			}
			catch(FileNotFoundException e) {
				System.err.println("The file " + options.bookMarkPath + " could not be found");
				return(false);
			}
			if(options.quiet == false) {
				System.out.println("Processing bookmark file: " + options.bookMarkPath);
			}

			String line;
			while((line = finb.readLine()) != null) {
				if(line.length() > 0 && line.charAt(0) != '!') {
					processBookMark(uncompressedText, line, bookMarks, options);
				}
			}
			finb.close();
		}

		///////////////////////////////////////////////////
		// auto-bookmark if specified
		if(options.autoBookMark) {
			char curChar;
			int startPos = -1;
			int endPos = -1;
			for(int curCharPos = uncompressedText.length() - 1; curCharPos > -1; curCharPos--) {
				curChar = uncompressedText.charAt(curCharPos);
				if(curChar == '<') {	// start sequence
					startPos = curCharPos;
					break;	// we are done looking
				}
				else if(curChar == '>') {	// end sequence
					endPos = curCharPos;
				}
				else if(!Character.isWhitespace(curChar) && endPos == -1) {
					break;
				}
			}
			if(startPos > -1 && endPos > -1) {
				String lookFor = uncompressedText.substring(startPos + 1, endPos);
				if(options.quiet == false) {
					System.out.println("Auto-bookmark: Looking for '" + lookFor + "'");
				}
				int found = 0;
				while((found = uncompressedText.indexOf(lookFor, found)) > -1) {
					if(found == 0 || uncompressedText.charAt(found - 1) == '\n') {
						BookMark bm = new BookMark();
						bm.fileOffset = found;
						int curCharPos = found += lookFor.length();
						for(; curCharPos < uncompressedText.length(); curCharPos++) {
							if(uncompressedText.charAt(curCharPos) == '\n') {
								while(Character.isWhitespace(uncompressedText.charAt(found))) {
									found++;
								}
								bm.name = uncompressedText.substring(found, curCharPos);
								if(options.quiet == false) {
									System.out.println("Found the '" + bm.name + "' bookmark");
								}
								bookMarks.add(bm);
								break;
							}
						}
						found = curCharPos;
					}
					else {
						found++;
					}
				}
			}
		}

		///////////////////////////////////////////////////
		// build the PalmDOC
		DatabaseHeader dbHeader = new DatabaseHeader();
		dbHeader.name = options.title;
		dbHeader.setModificationDate();
		dbHeader.creatorID = options.docID;
		dbHeader.typeID = DatabaseHeader.TEXt;

		docHeader.version = (options.compress ? DocumentHeader.COMPRESSED : DocumentHeader.UNCOMPRESSED);
		docHeader.recordSize = DocumentHeader.textRecordSize;

		docHeader.numRecords = (short)(docHeader.storyLen / DocumentHeader.textRecordSize);
		if(docHeader.numRecords * DocumentHeader.textRecordSize < docHeader.storyLen) {
			docHeader.numRecords++;
		}

		if(bookMarks.isEmpty() == false) {
			dbHeader.numRecords = (short)(docHeader.numRecords + 1 + bookMarks.size());
		}
		else {
			dbHeader.numRecords = (short)(docHeader.numRecords + 1);
		}
		dbHeader.write(fout);

		int bookMarkIndex = docHeader.numRecords + 1;	// this will only be used if we are bookmarking

		RecordIndex[] ri = new RecordIndex[dbHeader.numRecords];
		ri[0] = new RecordIndex();
		ri[0].attribute = (byte)(attribute | options.category);
		ri[0].write(fout);
		for(int i = 1; i < dbHeader.numRecords; i++) {
			ri[i] = new RecordIndex();
			ri[i].fileOffset = 0;
			ri[i].write(fout);
		}

		ri[0].fileOffset = (int) fout.getFilePointer();

		docHeader.write(fout);

		int totalCompressedBytes = 0;
		int processed = 0;
		int compressed = 0;
		int original = 0;

		for(int recNum = 1; recNum <= docHeader.numRecords; recNum++) {
			ri[recNum].fileOffset = (int)fout.getFilePointer();
			original = docHeader.recordSize;
			if(original + processed > uncompressedBuffer.length) {
				original = uncompressedBuffer.length - processed;
			}
			textBuffer.assign(uncompressedBuffer, original, processed);
			processed += original;
			if(options.compress) {
				compressed = textBuffer.compress();
				totalCompressedBytes += compressed;
			}
			fout.write(textBuffer.buf);
			if(options.quiet == false) {
				System.out.print("Converting record " + recNum + " of " + docHeader.numRecords);
				if(processed > 0 && options.compress) {
					System.out.print(" original " + original + " compressed to "
									 + compressed + " saved: " + (int)(100 - (100. * compressed / original + .5)) + "%");
				}
				System.out.println();
			}
		}

		// process the bookmarks
		if(bookMarks.isEmpty() == false) {
			if(options.sortBookMarks == 1) {
				Collections.sort(bookMarks, new SortBookMarksByLocation());
			}
			else if(options.sortBookMarks == 2) {
				Collections.sort(bookMarks, new SortBookMarksAlphabetically());
			}
			for(int bm = 0; bm < bookMarks.size(); bm++) {
				if(((BookMark)bookMarks.get(bm)).fileOffset == -1) {
					System.err.println("Bookmark " + ((BookMark)bookMarks.get(bm)).name + " not found");
					deletePilotDoc = true;
				}
				else {
					ri[bookMarkIndex + bm].fileOffset = (int)fout.getFilePointer();
					((BookMark)bookMarks.get(bm)).write(fout);
				}
			}
			if(options.quiet == false && deletePilotDoc == false) {
				System.out.println("Added " + bookMarks.size() + " bookmark(s)");
			}
		}

		// re-write the record index array, now with the correct file offsets
		fout.seek(DatabaseHeader.getSize());
		for(int i = 0; i < ri.length; i++) {
			ri[i].write(fout);
		}
		fout.close();

		if(options.quiet == false && options.compress) {
			System.out.println("Processed " + docHeader.storyLen +
							   " bytes, compressed to " + totalCompressedBytes +
							   " bytes, saved: " + (int)(100 - (100. * totalCompressedBytes / docHeader.storyLen + .5)) + "%");
		}

		if(deletePilotDoc) {
			System.err.println(">>>>>" + options.destinationFile + " is being DELETED<<<<<");
			options.destinationFile.delete();
			return(false);
		}
		Date stopTime = new Date();
		System.out.println("Encode time = " + (stopTime.getTime() - startTime.getTime()) + "ms");
		return(true);
	}

	private static void processBookMark(String uncompressedText, String line, ArrayList bookMarks, EncodeOptions options) {
		BookMarkHelperMD bmh = null;
		int startBMing = 0;
		int endBMing = Integer.MAX_VALUE;
		StringTokenizer lineToken = new StringTokenizer(line, ",");
		bmh = new BookMarkHelperMD();
		bmh.name = lineToken.nextToken();
		bmh.trimBmWhitespace=options.trimBmWhitespace;

		try {
			bmh.setSearchText(lineToken.nextToken());
			try {
				bmh.fileOffset = Integer.decode(bmh.getSearchText()).intValue();
			}
			catch(NumberFormatException e) {
				// was the search string
			}

			bmh.setReplaceText(lineToken.nextToken());
			try {
				startBMing = Integer.decode(bmh.getReplaceText()).intValue();
				bmh.setReplaceText(null);
			}
			catch(NumberFormatException e) {
				// was the replace string
			}

			try {
				if(bmh.getReplaceText() == null) {
					endBMing = Integer.decode(lineToken.nextToken()).intValue();
				}
				else {
					startBMing = Integer.decode(lineToken.nextToken()).intValue();
					endBMing = Integer.decode(lineToken.nextToken()).intValue();
				}
			}
			catch(NumberFormatException e) {
			}
		}
		catch(NoSuchElementException e) {
			// finished parsing bookmark specification line
		}

		if(bmh.fileOffset == -1) {
			// a string was given to search for
			int bmkFound = 0;
			int bmkAdded = 0;
			int progressIndex = 0;
			do {
				if(options.quiet == false && fromGUI == false) {
					System.out.print("\rsearching " + progress[progressIndex]);
					progressIndex++;
					if(progressIndex >= progress.length) {
						progressIndex = 0;
					}
				}
				// Start the search only one character after the last hit.
				// In the beginning fileOffset is -1, so we start at index 0.
				if(bmh.search(uncompressedText, bmh.endOfMatch)) {
					bmkFound++;
					if(bmkFound >= startBMing && bmkFound <= endBMing) {
						bookMarks.add(new BookMark(bmh));
						bmkAdded++;
					}
				}
			} while(bmh.ok2continue() || bmkFound > endBMing);
			if(options.quiet == false && fromGUI == false) {
				System.out.print("\r");
			}
			if(bmkAdded == 0) {
				if(bmh.getSearchText() == null) {
					System.out.println("Cannot find the `"
									   + bmh.name
									   + "` bookmark"
									  );
				}
				else {
					System.out.println("Cannot find the `"
									   + bmh.getSearchText()
									   + "` bookmark"
									  );
				}
			}
			if(options.quiet == false) {
				if(bmkAdded > 0) {
					if(bmh.getSearchText() == null) {
						System.out.println("Found the `"
										   + bmh.name
										   + "` bookmark "
										   + bmkFound
										   + " time(s), added "
										   + bmkAdded
										   + " time(s)."
										  );
					}
					else {
						System.out.println("Found the `"
										   + bmh.getSearchText()
										   + "` bookmark "
										   + bmkFound
										   + " time(s), added "
										   + bmkAdded
										   + " time(s)."
										  );
					}
				}
				else {
					if(bmh.getSearchText() == null) {
						System.out.println("Could not find `"
										   + bmh.name
										   + "` ");
					}
					else {
						System.out.println("Could not find `"
										   + bmh.getSearchText()
										   + "` ");
					}
				}
			}
		}
		else {
			// An integer is given as the string to search for,
			// so it indicates a position in the text.
			bookMarks.add(new BookMark(bmh));
			if(options.quiet == false) {
				System.out.println("Added the bookmark '"
								   + bmh.name
								   + "' at position "
								   + bmh.fileOffset);
			}
		}
	}

	/**
	 * Start of the application
	 * @param args arguments to MakeDocJ application
	 * @since 1.0
	 */
	public static void main(String args[]) {
/*
		System.out.println("================================================================================");
		System.out.println("main:Command line arguments");
		for(int arg = 0; arg < args.length; arg++) {
			System.out.println("args[" + arg + "]=" + args[arg]);
		}
		System.out.println();
		System.out.println("================================================================================");
*/
		System.out.println("MakeDocJ v" + version);
		if(args.length > 0) {
			if(args[0].charAt(0) == '@') {
				String optionFile = args[0].substring(1);
				try {
					LineNumberReader optFile = new LineNumberReader(new FileReader(optionFile));
					String line;
					while((line = optFile.readLine()) != null) {
						if(line.length() > 0 && line.charAt(0) != '!') {
							processLine(line, true);
							System.out.println();
						}
					}
				}
				catch(FileNotFoundException e) {
					System.err.println("the file " + optionFile + " could not be found");
					return;
				}
				catch(IOException e) {
					System.err.println("the file " + optionFile + " could not be processed");
					return;
				}
			}
			else {
				processArgs(args, true);
			}
		}
		else {
			usage(true);
		}
	}

	/**
	 * Processes the line of text passed into a command line type argument array,
	 * then calls processArgs()
	 * @param line The line to parse and process
	 * @param makeDocJ Indicates whether called from MakeDocJ or not.
	 * @since 3.8
	 */
	public static void processLine(String line, boolean makeDocJ) {
		ArrayList arguments = new ArrayList();
		int tokens = 0;
		boolean inQuoteArgument = false;
		boolean removeEndQuote = false;
		char curChar;
		int startToken = 0;
		for(int curCharPos = 0; curCharPos < line.length(); curCharPos++) {
			curChar = line.charAt(curCharPos);
			if(curChar == '"') {
				if(inQuoteArgument) {
					inQuoteArgument = false;
				}
				else {
					inQuoteArgument = true;
					startToken++;	// So we don't get quote
					removeEndQuote = true;
				}
			}
			if(curChar == ' ' && inQuoteArgument == false) {
				if(removeEndQuote) {
					removeEndQuote = false;
					arguments.add(line.substring(startToken, curCharPos - 1));
				}
				else {
					arguments.add(line.substring(startToken, curCharPos));
				}
				startToken = curCharPos + 1;
			}
		}
		if(inQuoteArgument) {
			System.err.println("Unmatching quotes in command line");
			return;
		}
		else {
			if(removeEndQuote) {
				arguments.add(line.substring(startToken, line.length() - 1));
			}
			else {
				arguments.add(line.substring(startToken, line.length()));
			}
		}

		String optArgs[] = new String[arguments.size()];
		for(int curToken = 0; curToken < arguments.size(); curToken++) {
			optArgs[curToken] = (String)arguments.get(curToken);
		}
		processArgs(optArgs, makeDocJ);
	}

	/**
	 * processes argument array
	 * @param args arguments to MakeDocJ application
	 * @param makeDocJ Indicates whether called from MakeDocJ or not.
	 * @since 2.1
	 */
	private static void processArgs(String args[], boolean makeDocJ) {
/*
		System.out.println("================================================================================");
		System.out.println("processArgs:Command line arguments");
		for(int arg = 0; arg < args.length; arg++) {
			System.out.println("args[" + arg + "]=" + args[arg]);
		}
		System.out.println();
		System.out.println("================================================================================");
*/
		if(args.length < 3) {
			usage(makeDocJ);
		}
		else {
			int argNum = -1;
			boolean decode = false;	// default to compress
			EncodeOptions eOptions = new EncodeOptions();
			DecodeOptions dOptions = new DecodeOptions();
			Options options = eOptions;

			while(argNum++ < args.length && args[argNum].charAt(0) == '-') {
				if(args[argNum].charAt(1) == '0') {		// OS dependant
					dOptions.eolType = PilotDocRecord.EOLVM;
					continue;
				}
				if(args[argNum].charAt(1) == '1') {		// DOS => CR/LF
					dOptions.eolType = PilotDocRecord.EOLCRLF;
					continue;
				}
				if(args[argNum].charAt(1) == '2') {		// MAC => CR
					dOptions.eolType = PilotDocRecord.EOLCR;
					continue;
				}
				if(args[argNum].charAt(1) == 'a') {
					eOptions.autoBookMark = true;
					continue;
				}
				if(args[argNum].charAt(1) == 'b') {
					dOptions.bookmarks = true;
					continue;
				}
				if(args[argNum].charAt(1) == 'B') {
					argNum++;
					eOptions.bookMarks.add(args[argNum]);
					continue;
				}
				if(args[argNum].charAt(1) == 'c') {
					argNum++;
					eOptions.category = Byte.parseByte(args[argNum]);
					continue;
				}
				if(args[argNum].charAt(1) == 'd') {
					decode = true;
					options = dOptions;
					continue;
				}
				if(args[argNum].charAt(1) == 'i') {
					argNum++;
					if(args[argNum].length() != 4) {
						System.err.println(args[argNum] + " is incorrect length, must be 4 characters");
						System.exit(1);
					}
					else {
						eOptions.docID = DatabaseHeader.convertStringToCreatorID(args[argNum]);
					}
					continue;
				}
				if(args[argNum].charAt(1) == 'm') {
					eOptions.compressEOL = true;
					continue;
				}
				if(args[argNum].charAt(1) == 'n') {
					eOptions.compress = false;
					continue;
				}
				if(args[argNum].charAt(1) == 'p') {
					eOptions.makePrivate = true;
					continue;
				}
				if(args[argNum].charAt(1) == 'q') {
					options.quiet = true;
					continue;
				}
				if(args[argNum].charAt(1) == 'r') {
					eOptions.binary = false;
					continue;
				}
				if(args[argNum].charAt(1) == 's') {
					argNum++;
					eOptions.sortBookMarks = Integer.parseInt(args[argNum]);
					continue;
				}
				if(args[argNum].charAt(1) == 'u') {
					eOptions.unwrap = true;
					try {
						argNum++;
						eOptions.lineLength = Integer.parseInt(args[argNum]);
					}
					catch(NumberFormatException e) {
						argNum--;
					}
					continue;
				}
				if(args[argNum].charAt(1) == 'w') {
					eOptions.trimBmWhitespace = true;
					continue;
				}
				if(args[argNum].charAt(1) == '?') {
					usage(makeDocJ);
					return;
				}
				System.err.println("Unknown option `" + args[argNum].charAt(1) + "`");
				usage(makeDocJ);
				return;
			}

			options.sourcePath = args[argNum];
			argNum++;

			options.destinationFile = new File(args[argNum]);
			argNum++;

			try {
				if(decode) {
					decode(dOptions);
				}
				else {
					eOptions.title = args[argNum];
					argNum++;

					try {
						eOptions.bookMarkPath = new File(args[argNum]);
						argNum++;
					}
					catch(java.lang.ArrayIndexOutOfBoundsException e) {
					}

					encode(eOptions);
				}
			}
			catch(IOException e) {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Displays usage
	 * @since 1.0
	 * @param makeDocJ Indicates whether called from MakeDocJ or not.
	 */
	private static void usage(boolean makeDocJ) {
		if(makeDocJ) {
			System.out.println("Usage: MakeDocJ [-options] ...");
			System.out.println();
		}
		System.out.println("To encode text file into PilotDOC format file");
		System.out.println();
		if(makeDocJ) {
			System.out.print("MakeDocJ ");
		}
		System.out.println("[-options] <text-file> <PilotDOC-file> <story-name> [<bookmark-file>]");
		System.out.println("    -a auto-bookmark (In addition to other bookmark methods)");
		System.out.println("    -c {number} category number of document");
		System.out.println("    -i {database ID} four character database ID");
		System.out.println("    -m compress multi-linefeeds out of <text-file>");
		System.out.println("    -n builds the PilotDOC file without compression");
		System.out.println("    -p sets the PilotDoc to private");
		System.out.println("    -q quiet (i.e. minimal reporting during encoding process)");
		System.out.println("    -r remove binary");
		System.out.println("    -s {number} sort bookmarks {number}: 1 = By location, 2 = Alphabetically");
		System.out.println("    -u {number} unwrap text, single LF is removed");
		System.out.println("       If {number} is specified, line length must be greater than {number}");
		System.out.println("       to wrap, default: 0");
		System.out.println("    -w remove spaces at the beginning of the bookmark text");
		System.out.println("    -B {bookmark} (In addition to other bookmark methods)");
		System.out.println("       Multiple of these are allow to specify more that one bookmark");
		System.out.println();
		System.out.println();
		System.out.println("To decode PilotDOC format file into the text file");
		System.out.println();
		if(makeDocJ) {
			System.out.print("MakeDocJ ");
		}
		System.out.println("-d [-options] <PilotDOC-file> <text-file>");
		System.out.println("    -0 end of line is based on OS");
		System.out.println("    -1 end of line is CR/LF (i.e. DOS)");
		System.out.println("    -2 end of line is CR    (i.e. MAC)");
		System.out.println("    -b generate a bookmark file (<text-file>.bm) if there are any bookmarks");
		System.out.println("    -d decodes the PilotDOC file into a text file");
		System.out.println("    -q quiet (i.e. minimal reporting during decoding process)");
		if(makeDocJ) {
			System.out.println();
			System.out.println("MakeDocJ @<options-file>");
			System.out.println("  process <options-file> to get options");
		}
	}

	// called from a GUI
	static boolean fromGUI = false;
}

/**
 * <code>BookMarkHelperMD</code> is used to help generate bookmarks in the PilotDoc
 * @author Jeffrey A. Krzysztow
 * @author Reinhold F. Kainhofer
 * @version 1.3
 */
class BookMarkHelperMD extends BookMark {

	/**
	 * sets the regular expression text we will use in search
	 * @since 1.2
	 */
	final public void setSearchText(String searchText) {
		this.searchText = searchText;
		searchRE = null;
	}

	/**
	 * returns the regular expression search text
	 * @return the search text
	 * @since 1.2
	 */
	final public String getSearchText() {
		return(searchText);
	}

	/**
	 * sets the regular expression text we will use to generate the bookmark name
	 * @since 1.3
	 */
	final public void setReplaceText(String replaceText) {
		this.replaceText = replaceText;
	}

	/**
	 * returns the regular expression text we will use to generate the bookmark name
	 * @return the replace text
	 * @since 1.3
	 */
	final public String getReplaceText() {
		return(replaceText);
	}

	private RE searchRE = null;
	private RESyntax syntax = null;
	private String searchText = null;
	private String replaceText = null;
	private boolean autoSetName = false;
	private boolean multiMatch = false;
	private boolean lastSearchSuccessful = false;
	public boolean trimBmWhitespace = false;
	public int endOfMatch = 0;

	/**
	 * searches text2Search for the seachText and sets the fileOffset from BookMark
	 * @param text2Search text to search
	 * @param index the offset where to start the search in the string
	 * @return whether it found the text or not
	 * @since 1.0
	 */
	public boolean search(String text2Search, int index) {
		lastSearchSuccessful = false;
		if(searchText == null) {
			fileOffset = text2Search.indexOf(name, index);
			if(fileOffset > -1) {
				lastSearchSuccessful = true;
				endOfMatch = fileOffset + name.length();
			}
		}
		else {
			try {
				try {
					if(searchRE == null) {
						syntax = new RESyntax(RESyntax.RE_SYNTAX_PERL5);
						syntax.setLineSeparator("\n");
						searchRE = new RE(searchText, RE.REG_MULTILINE, syntax);
						if(name.length() == 1) {
							switch(name.charAt(0)) {
								case '-':
									autoSetName = true;
									multiMatch = false;
									break;
								case '+':
									autoSetName = true;
									multiMatch = true;
									break;
							}
						}
					}
					REMatch rem = searchRE.getMatch(text2Search, index);
					if(null != rem) {
						lastSearchSuccessful = true;
						fileOffset = rem.getStartIndex();
						endOfMatch = rem.getEndIndex();
						if(autoSetName) {
							if(null == replaceText) {
								name = rem.toString();
								if(trimBmWhitespace) {
									name = name.trim();
								}
							}
							else {
								name = rem.substituteInto(replaceText);
							}
						}
					}
					else {
						fileOffset = -1;
					}
				}
				catch(IllegalArgumentException e) {
					e.printStackTrace();
				}
			}
			catch(REException e) {
				e.printStackTrace();
			}
		}
		return(lastSearchSuccessful);
	}

	/**
	 * Determine if we have more matching to do.
	 * @return whether we have more matching to do.
	 * @since 1.0
	 */
	final public boolean ok2continue() {
		return(lastSearchSuccessful && multiMatch);
	}

	/**
	 * override Object.toString()
	 * @since 1.0
	 */
	public String toString() {
		return(">> BookMarkHelperMD <<\nname = `" + name
			   + "`\nfileOffset = " + fileOffset
			   + "\nsearchText = " + searchText);
	}
}

class Options implements Cloneable {

	public Object clone() {
		Options options = new Options();
		clone(options);
		return(options);
	}

	protected void clone(Options options) {
		try {
			options.sourcePath = sourcePath;
		}
		catch(Exception e) {
			options.sourcePath = null;
		}
		try {
			options.destinationFile = destinationFile;
		}
		catch(Exception e) {
			options.destinationFile = null;
		}
		options.quiet = quiet;
	}

	// full path of source text file
	String sourcePath = null;

	// full path of destination PilotDoc file
	File destinationFile = null;

	// don't print as much information while processing
	boolean quiet = false;
}

class EncodeOptions extends Options implements Cloneable {

	public Object clone() {
		EncodeOptions options = new EncodeOptions();
		clone(options);
		return(options);
	}

	protected void clone(EncodeOptions options) {
		super.clone(options);
		try {
			options.bookMarkPath = bookMarkPath;
		}
		catch(Exception e) {
			options.bookMarkPath = null;
		}
		if(title != null) {
			options.title = new String(title);
		}
		options.compress = compress;
		options.binary = binary;
		options.docID = docID;
		options.compressEOL = compressEOL;
		options.unwrap = unwrap;
		options.lineLength = lineLength;
		options.makePrivate = makePrivate;
		options.category = category;
		options.sortBookMarks = sortBookMarks;
		options.autoBookMark = autoBookMark;
		options.trimBmWhitespace = trimBmWhitespace;
		options.bookMarks.addAll(bookMarks);
	}

	// full path of bookmark file
	File bookMarkPath = null;

	// title to embed in PilotDoc
	String title = null;

	// compress text in PilotDoc
	boolean compress = true;

	// process as binary file
	boolean binary = false;

	// PilotDoc ID
	// 'REAd' database creator ID, generic reader
	int docID = DatabaseHeader.ReaderID;

	// remove extra end of line characters
	boolean compressEOL = false;

	// "unwrap" text
	boolean unwrap = false;

	// only unwrap if line is longer than lineLength
	int lineLength = 0;

	// make to PilotDoc private
	boolean makePrivate = false;

	// the PilotDoc category
	byte category = 0;

	// sort the bookmarks
	// 0 = none
	// 1 = by location
	// 2 = alphabetically
	int sortBookMarks = 0;

	// auto-bookmark - generate bookmarks from the <> standard
	boolean autoBookMark = false;

	// trim whitespace from both ends of the bookmark string.
	// useful if headline have a huge indentation to make them appear centered
	boolean trimBmWhitespace = false;

	// command line specified bookmarks
	ArrayList bookMarks = new ArrayList();
}

class DecodeOptions extends Options implements Cloneable {

	public Object clone() {
		DecodeOptions options = new DecodeOptions();
		clone(options);
		return(options);
	}

	protected void clone(DecodeOptions options) {
		super.clone(options);
		options.bookmarks = bookmarks;
		options.eolType = eolType;
	}

	// generate a file containing the internal bookmarks
	boolean bookmarks = false;

	// how to convert end of line
	int eolType = PilotDocRecord.EOLLF;	// UNIX => LF
}

class SortBookMarksByLocation implements Comparator {
	/**
	 * Compares its two arguments for order.  Returns a negative integer,
	 * zero, or a positive integer as the first argument is less than, equal
	 * to, or greater than the second.<p>
	 *
	 * The implementor must ensure that <tt>sgn(compare(x, y)) ==
	 * -sgn(compare(y, x))</tt> for all <tt>x</tt> and <tt>y</tt>.  (This
	 * implies that <tt>compare(x, y)</tt> must throw an exception if and only
	 * if <tt>compare(y, x)</tt> throws an exception.)<p>
	 *
	 * The implementor must also ensure that the relation is transitive:
	 * <tt>((compare(x, y)&gt;0) &amp;&amp; (compare(y, z)&gt;0))</tt> implies
	 * <tt>compare(x, z)&gt;0</tt>.<p>
	 *
	 * Finally, the implementer must ensure that <tt>compare(x, y)==0</tt>
	 * implies that <tt>sgn(compare(x, z))==sgn(compare(y, z))</tt> for all
	 * <tt>z</tt>.<p>
	 *
	 * It is generally the case, but <i>not</i> strictly required that
	 * <tt>(compare(x, y)==0) == (x.equals(y))</tt>.  Generally speaking,
	 * any comparator that violates this condition should clearly indicate
	 * this fact.  The recommended language is "Note: this comparator
	 * imposes orderings that are inconsistent with equals."
	 *
	 * @param o1 the first object to be compared.
	 * @param o2 the second object to be compared.
	 * @return a negative integer, zero, or a positive integer as the
	 * 	       first argument is less than, equal to, or greater than the
	 *	       second.
	 * @throws ClassCastException if the arguments' types prevent them from
	 * 	       being compared by this Comparator.
	 */
	public int compare(Object o1, Object o2) {
		return(
			  ((BookMark)o1).fileOffset < ((BookMark)o2).fileOffset ? -1 :
			  ((BookMark)o1).fileOffset == ((BookMark)o2).fileOffset ? 0 : 1);
	}
}

class SortBookMarksAlphabetically implements Comparator {
	/**
	 * Compares its two arguments for order.  Returns a negative integer,
	 * zero, or a positive integer as the first argument is less than, equal
	 * to, or greater than the second.<p>
	 *
	 * The implementor must ensure that <tt>sgn(compare(x, y)) ==
	 * -sgn(compare(y, x))</tt> for all <tt>x</tt> and <tt>y</tt>.  (This
	 * implies that <tt>compare(x, y)</tt> must throw an exception if and only
	 * if <tt>compare(y, x)</tt> throws an exception.)<p>
	 *
	 * The implementor must also ensure that the relation is transitive:
	 * <tt>((compare(x, y)&gt;0) &amp;&amp; (compare(y, z)&gt;0))</tt> implies
	 * <tt>compare(x, z)&gt;0</tt>.<p>
	 *
	 * Finally, the implementer must ensure that <tt>compare(x, y)==0</tt>
	 * implies that <tt>sgn(compare(x, z))==sgn(compare(y, z))</tt> for all
	 * <tt>z</tt>.<p>
	 *
	 * It is generally the case, but <i>not</i> strictly required that
	 * <tt>(compare(x, y)==0) == (x.equals(y))</tt>.  Generally speaking,
	 * any comparator that violates this condition should clearly indicate
	 * this fact.  The recommended language is "Note: this comparator
	 * imposes orderings that are inconsistent with equals."
	 *
	 * @param o1 the first object to be compared.
	 * @param o2 the second object to be compared.
	 * @return a negative integer, zero, or a positive integer as the
	 * 	       first argument is less than, equal to, or greater than the
	 *	       second.
	 * @throws ClassCastException if the arguments' types prevent them from
	 * 	       being compared by this Comparator.
	 */
	public int compare(Object o1, Object o2) {
		return((BookMark)o1).name.compareTo(((BookMark)o2).name);

	}
}

