/*
 * PilotDocRecord.java
 *
 * Copyright 2003 by BRiSK Software,
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: PilotDocRecord.java,v 1.6 2006/02/08 05:03:06 brisk Exp $
 */

/**
 * <code>PilotDocRecord</code> class handles PilotDoc records. It can
 * compress, decompress, remove binary codes, duplicate \n codes
 *
 * @author Jeffrey A. Krzysztow
 * @author Pat Beirne
 * @version 1.5.1
 */

import java.io.*;

// NOTE: for JDK 1.1.x get special collections
// from Sun http://java.sun.com/products/javabeans/infobus/collectionsreadme.html
// import com.sun.java.util.collections.*;
// NOTE: for JDK 1.2+
// import java.util.*;
import @COLLECTIONS@; // used by ant

public class PilotDocRecord {
	final static int DISP_BITS = 11;
	final static int COUNT_BITS = 3;
	final static int DEF_LEN = DocumentHeader.textRecordSize;
	final static int SHIFTED_DISP_BITS_MINUS_1 = ((1 << DISP_BITS) - 1);
	final static int SHIFTED_COUNT_BITS_MINUS_1 = ((1 << COUNT_BITS) - 1);
	final static int SHIFTED_COUNT_BITS_PLUS_2 = (1 << COUNT_BITS) + 2;

	final static int EOLVM = 0;	// ask VM for default
	final static int EOLLF = 1;
	final static int EOLCR = 2;
	final static int EOLCRLF = 3;

	/**
	 * internal buffer holding PilotDocRecord
	 * @since 1.0
	 */
	byte [] buf = null;

	// used during compress, decompress, this it the destination buffer
	private byte [] destBuf = null;
	private int destIndex = 0;
	private boolean space = false;

	/**
	 * Constructs a PilotDocRecord of default size
	 * @since 1.0
	 */
	PilotDocRecord() {
		this(DEF_LEN);
	}

	/**
	 * Constructs a PilotDocRecord of the specified size
	 * @param size size to create PilotDocRecord
	 * @since 1.0
	 */
	PilotDocRecord(int size) {
		buf = new byte[size];
		setDestBuf(size);
	}

	/**
	 * Sets the destination buffer for 
	 * @param size size to create the destination buffer
	 */
	private void setDestBuf(int size) {
		if(size > DocumentHeader.textRecordSize) {
			size = size * 2;
		}
		else {
			size = DocumentHeader.textRecordSize * 2;
		}
		destBuf = new byte[size * 2]; // this should be overkill
	}

	/**
	 * Returns the number of bytes in the PilotDocRecord
	 * @return the number of bytes in the PilotDocRecord
	 * @since 1.0
	 */
	public int length() {
		return buf.length;
	}

	/**
	 * Remove the binary codes throw away really low ASCII or above '~' from the buffer
	 * @return the resulting buffer length
	 * @since 1.0
	 */
	public int removeBinary() {
		for(int bufIndex = destIndex = 0; bufIndex < buf.length; bufIndex++, destIndex++) {
			// copy each byte
			destBuf[destIndex] = buf[bufIndex];
			
			// throw away really low ASCII
			if((destBuf[destIndex] >= 0x00 && destBuf[bufIndex] < 0x09)) {
				destIndex--;
			}
		}
		assign(destBuf, destIndex);
		return destIndex;
	}

	/**
	 * Change all 0x0a into 0x0d/0x0a pairs or 0x0a into 0x0d
	 * @return the resulting buffer length
	 * @since 1.0
	 */
	public int convertEOL(final int eolType) {
		byte [] lineSepBytes = null;

		if(eolType == EOLVM) {
			String lineSeparator = System.getProperty("line.separator"); // remeber setting
			lineSepBytes = lineSeparator.getBytes();
		}
		else if(eolType == EOLCR) {
			lineSepBytes = new byte [] { '\r' };
		}
		else if(eolType == EOLLF) {
			lineSepBytes = new byte [] { '\n' };
		}
		else if(eolType == EOLCRLF) {
			lineSepBytes = new byte [] { '\r', '\n' };
		}
		for(int bufIndex = destIndex = 0; bufIndex < buf.length; bufIndex++, destIndex++) {
			destBuf[destIndex] = buf[bufIndex];
			if(destBuf[destIndex] == 0x0a) {
				int l = 0;
				for(; l < lineSepBytes.length; l++) {
					destBuf[destIndex + l] = lineSepBytes[l];
				}
				destIndex = destIndex + l - 1;
			}			
		}
		assign(destBuf, destIndex);
		return destIndex;
	}

	/**
	 * if a CR appears before a LF, then remove the CR
	 * if a CR appears in isolation, change to a LF
	 * @return the resulting buffer length
	 * @since 1.0
	 */
	public int convertEOL() {
		for(int bufIndex = destIndex = 0; bufIndex < buf.length; bufIndex++, destIndex++) {
			// copy each byte
			destBuf[destIndex] = buf[bufIndex];

			if(destBuf[destIndex] == 0x0d) {
				// if next is LF, then drop it
				if(bufIndex < buf.length - 1 && buf[bufIndex + 1] == 0x0a) {
					destIndex--;
				}
				else { // turn it into a LF
					destBuf[destIndex] = 0x0a;
				}
			}
		}
		assign(destBuf, destIndex);
		return destIndex;
	}
	
	/**
	 * unwrap single LFs are removed
	 * @param the minimum line length to wrap
	 * @return the resulting buffer length
	 * @since 1.0
	 */
	public int unwrap(int lineLength) {
		int lastLFpos = -1, x, y;
		
		for(int bufIndex = destIndex = 0; bufIndex < buf.length; bufIndex++, destIndex++) {
			destBuf[destIndex] = buf[bufIndex];
			
			if(destBuf[destIndex] == 0x0a) {
				// Test for some special text patterns
				boolean patternFound = false;

				if(patternFound == false) { // have not found a pattern, so look for one
					patternFound = true; // be optimistic
					// test to see if mirror of each other
					// test to see if all bytes are the same
					for(x = lastLFpos + 1, y = destIndex - 1; x < y; x++, y--) {
						if(destBuf[x] != destBuf[y]){
							patternFound = false;
							break;
						}
					}
					if(patternFound) {
						if(lastLFpos > -1) {
							destBuf[lastLFpos] = 0x0a; // put the LF back so we have a "complete" line
						}
					}
				}
				
				if(patternFound == false) { // unwrap
					if ((destIndex - lastLFpos) > lineLength) {
						if(destBuf[destIndex - 1] == 0x20 || destBuf[destIndex - 1] == 0x0a) {
							destIndex--;
						}
						else {
							destBuf[destIndex] = 0x20;
						}
					}
				}
				lastLFpos = destIndex;
			}
		}
		assign(destBuf, destIndex);
		return destIndex;
	}

	/**
	 * compressEOL compresses double LF into single LF
	 * @return the resulting buffer length
	 * @since 1.3
	 */
	public int compressEOL() {
		for(int bufIndex = destIndex = 0; bufIndex < buf.length; bufIndex++, destIndex++) {
			// copy each byte
			destBuf[destIndex] = buf[bufIndex];

			// throw away one of the LF
			if(destIndex > 0 && destBuf[destIndex] == 0x0a && destBuf[destIndex - 1] == 0x0a) {
				destIndex--;
			}
		}
		assign(destBuf, destIndex);
		return destIndex;
	}
	
	/**
	 * Decompress the buffer
	 * action: make a new buffer
	 *				run through the source data
	 *				check the 4 cases:
	 *					0,9...7F represent self
	 *					1...8		escape n chars
	 *					80...bf reference earlier run
	 *					c0...ff	space+ASCII
	 * @return the resulting buffer length
	 * @since 1.0
	 */
	public int decompress() {
		for(int bufIndex = destIndex = 0; bufIndex < buf.length;) {
			// take a char from the input buffer
			int c = buf[bufIndex++] & 0xff;
			// separate the char into zones: 0, 1...8, 9...0x7F, 0x80...0xBF, 0xC0...0xFF
			// codes 1...8 mean copy that many bytes; for accented chars & binary
			if(c > 0 && c < 9) {
				System.arraycopy(buf, bufIndex, destBuf, destIndex, c);
				bufIndex += c;
			}
			// codes 0, 9...0x7F represent themselves
			else if(c < 0x80) {
				destBuf[destIndex++] = (byte) c;
			}

			// codes 0xC0...0xFF represent "space + ascii char"
			else if(c >= 0xC0) {
				destBuf[destIndex++] = (byte) ' ';
				destBuf[destIndex++] = (byte) (c ^ 0x80);
			}

			// codes 0x80...0xBf represent sequences
			else {
				c = (c << 8) | buf[bufIndex++] & 0xff;

				int distance = (c & 0x3fff) >> COUNT_BITS;
				int length = (c & SHIFTED_COUNT_BITS_MINUS_1) + 3;
				while(length-- > 0) {
					destBuf[destIndex] = destBuf[destIndex - distance];
					destIndex++;
				}
			}
		}
		assign(destBuf, destIndex);
		return destIndex;
	}

	/**
	 * compress the buffer
	 * @return the resulting buffer length
	 * @since 1.0
	 */
	public int compress() {
		int hit = 0;			// points to a walking test hit; works upwards on successive matches
		int prevHit = 0; 		// previous value of hit
		int testHead = 0;		// current test string
		int testTail = 1;		// current walking pointer; one past the current test buffer
		int end = buf.length;	// 1 past the end of the input buffer

		space = false;

		// loop, absorbing one more char from the input buffer on each pass
		for(destIndex = 0; testHead != end; testTail++) {
			if(testHead - prevHit > SHIFTED_DISP_BITS_MINUS_1) {
				prevHit = testHead - SHIFTED_DISP_BITS_MINUS_1;
			}

			// scan in the previous data for a match
//			hit = findPattern(testHead, testTail, prevHit);
			hit = findPatternBoyerMoore(testHead, testTail, prevHit);

			// on a mismatch or end of buffer, issue codes
			if(hit == -1
				|| hit == testHead
				|| testTail - testHead > SHIFTED_COUNT_BITS_PLUS_2
				|| testTail == end) {

				// first, check for short runs
				if(testTail - testHead < 4) {
					issue(buf[testHead]);
					testHead++;
				}

				// for inter runs, issue a run-code
				else {
					// issue space char if required
					if(space) {
						destBuf[destIndex++] = (byte) ' ';
						space = false;
					}

					int dist = testHead - prevHit;
					int compound = (dist << COUNT_BITS) + testTail - testHead - 4;

					destBuf[destIndex++] = (byte) (0x80 + (compound >> 8));
					destBuf[destIndex++] = (byte) (compound & 0xff);
					testHead = testTail - 1;
				}
				// start the search again
				prevHit = 0;
			}
			// got a match
			else {
				prevHit = hit;
			}
			// when we get to the end of the buffer,
			// don't increment past the end this forces the residue chars out one at a time
			if(testTail == end) {
				testTail--;
			}
		}

		// clean up any dangling spaces
		if(space) {
			destBuf[destIndex++] = (byte) ' ';
		}
		
		// final scan to merge consecutive high chars together
		int k, i, destTemp;
		for(i = k = 0; i < destIndex; i++, k++) {
			destBuf[k] = destBuf[i];
			destTemp = destBuf[k] & 0xff;
			// skip the run-length codes
			if(destTemp >= 0x80 && destTemp < 0xc0) {
				destBuf[++k] = destBuf[++i];
			}
			// if we hit a high char marker, look ahead for another
			else if(destBuf[k] == 1) {
				destBuf[k + 1] = destBuf[i + 1];
				while(i + 2 < destIndex && destBuf[i + 2] == 1 && destBuf[k] < 8) {
					destBuf[k]++;
					destBuf[k + destBuf[k]] = destBuf[i + 3];
					i += 2;
				}
				k += destBuf[k];
				i++;
			}
		}

		assign(destBuf, k);
		return k;
	}

	/**
	 * Finds the pattern specified by startPattern and endPattern within buf.
	 * Start searching within buf at startSearch.
	 * @param startPattern were the begining of the pattern is within buf
	 * @param endPattern were the end of the pattern is within buf
	 * @param startSerach were to start the search within buf
	 * @return the start of the pattern found or -1 if the pattern is not found.
	 * @since 1.5
	 */
	private int findPattern(int startPattern, int endPattern, int startSearch) {
		int patternLength = endPattern - startPattern;
		for(; startSearch < buf.length; startSearch++) {
			if(buf[startSearch] == buf[startPattern]) {
				int offset = 1;
				// located first match in sequence, see if the rest of the pattern matches
				for(; offset < patternLength; offset++) {
					if(buf[startSearch + offset] != buf[startPattern + offset]) {
						break;
					}
				}
				if(offset == patternLength) {
					return startSearch;
				}
			}
		}
		return -1;
	}

	/**
	 * Finds the pattern specified by startPattern and endPattern within buf.
	 * Start searching within buf at startSearch.
	 * @param startPattern were the begining of the pattern is within buf
	 * @param endPattern were the end of the pattern is within buf
	 * @param startSerach were to start the search within buf
	 * @return the start of the pattern found or -1 if the pattern is not found.
	 * @since 1.5.1
	 */
	private int findPatternBoyerMoore(int startPattern, int endPattern, int startSearch) {
		final int plen = endPattern - startPattern;
		if (plen == 0) return startSearch;

		// init the offset array:
		Arrays.fill(offsets, plen);		// default value is plen
		for (int skip=plen-1, sIdx=0; sIdx<plen; sIdx++, skip--) {
			offsets[buf[startPattern + sIdx]+128] = skip;
		}
		startSearch += plen;
		int searchIdx = plen;
		final int idxStart = plen-1;
		final int dlen = buf.length;
		do {
			startSearch--; searchIdx--;
			while (buf[startSearch] != buf[startPattern+searchIdx]) {
				final int t = offsets[buf[startSearch]+128];
				final int poff = plen-searchIdx;
				startSearch += (poff > t) ? poff : t;
				if (startSearch >= dlen) {
					return -1;		// not found
				}
				searchIdx = idxStart;
			}
		} 
		while (searchIdx > 0);
		return startSearch; 
	}
	
	private int [] offsets = new int[256];	
	
	/**
	 * Handle the details of writing a single character to the compressed stream used by compress()
	 * @since 1.0
	 */
	private void issue(byte src) {
		// if there is an outstanding space char, see if we can squeeze it in with an ASCII char
		int iSrc = src;
		if(space) {
			if(iSrc >= 0x40 && iSrc <= 0x7F) {
				destBuf[destIndex++] = (byte)(iSrc ^ 0x80);
			}
			else {
				// couldn't squeeze it in, so issue the space char by itself
				// most chars go out simple, except the range 1...8,0x80...0xFF
				destBuf[destIndex++] = 0x20;
				if(iSrc < 0x80 && (iSrc == 0 || iSrc > 8)) {
					destBuf[destIndex++] = src;
				}
				else {
					destBuf[destIndex++] = 1;
					destBuf[destIndex++] = src;
				}
			}
			// knock down the space flag
			space = false;
		}
		else {
			// check for a space char
			if(iSrc == 0x20) {
				space = true;
			}
			else {
				if(iSrc < 0x80 && (iSrc == 0 || iSrc > 8)) {
					destBuf[destIndex++] = src;
				}
				else {
					destBuf[destIndex++] = 1;
					destBuf[destIndex++] = src;
				}
			}
		}
	}

	/**
	 * sets the internal buffer to the byte array
	 * @param b byte array to to assign to the internal buffer
	 * @param n number of byte in b to use
	 * @since 1.0
	 */
	public void assign(byte[] b, int n) {
		buf = new byte[n];
		System.arraycopy(b, 0, buf, 0, n);
	}

	/**
	 * sets the internal buffer to the byte array
	 * @param b      byte array to to assign to the internal buffer
	 * @param n      number of byte in b to use
	 * @param offset starting location in b
	 * @since 1.0
	 */
	public void assign(byte[] b, int n, int offset) {
		buf = new byte[n];
		System.arraycopy(b, offset, buf, 0, n);
	}

}

