/*
 * INIFile.java
 *
 * Copyright 2003 by BRiSK Software,
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: INIFile.java,v 1.3 2003/05/23 21:44:31 brisk Exp $
 */

package com.BRiSK.INIFile;

import java.io.*;
import java.util.*;

/**
 * <code>INIFile</code> class provides the ability to manipulate .INI files
 * made popular by Microsoft Windows
 * @author Jeffrey A. Krzysztow
 * @version 2.1
 */
public class INIFile {
	/**
	 * constructor
	 * @since 2.1
	 */
	public INIFile() {
	}

	/**
	 * constructor
	 * @param the path to the INI file to be processing
	 * @since 1.0
	 */
	public INIFile(String iniFileName) {
		this.iniFile = new File(iniFileName);
		try {
			read();
		}
		catch(IOException e) {
		}
	}

	/**
	 * constructor
	 * @param the File of the INI file to be processing
	 * @since 1.0
	 */
	public INIFile(File iniFile) {
		this.iniFile = iniFile;
		try {
			read();
		}
		catch(IOException e) {
		}
	}

	/**
	 * reads the INI file specified in the constructor
	 * NOTE: comments in INI are ignored and lost when written.
	 * @throws IOException for any file IO errors
	 * @since 1.0
	 */
	public void read() throws IOException {
		clear();
		try {
			BufferedReader f = new BufferedReader(new FileReader(iniFile));
			String line = "";
			int equals = -1;
			int endBracket = -1;
			do {
				line = f.readLine();
				if(line != null) {
					line = line.trim();
					if(line.equals("") == false && line.startsWith(";") == false) {
						if(line.startsWith("[")) {
							if((endBracket = line.indexOf("]")) > -1) {
								line = line.substring(1, endBracket);
								currSection = new Section(line);
								sections.addElement(currSection);
							}
						}
						else if((equals = line.indexOf("=")) > -1) {
							if(currSection != null) {
								String keyName = line.substring(0, equals);
								String keyValue = line.substring(equals + 1);
								currSection.addKey(keyName, keyValue);
							}
						}
					}
				}
			} while(line != null);
			f.close();
		}
		catch(IOException e) {
			sections.removeAllElements();
			throw e;
		}
	}

	/**
	 * writes the INI file specified in the constructor
	 * NOTE: comments in INI are ignored and lost when written.
	 * @throws IOException for any file IO errors
	 * @since 1.0
	 */
	public void write() throws IOException {
		PrintWriter f = new PrintWriter(new FileOutputStream(iniFile));
		for(int s = 0; s < sections.size(); s++) {
			Section section = (Section)sections.elementAt(s);
			f.println();
			f.println("[" + section.name + "]");
			for(int k = 0; k < section.size(); k++) {
				Key key = section.elementAt(k);
				f.println(key.name + "=" + key.value);
			}
		}
		f.close();
	}

	/**
	 * writes the INI to the
	 * NOTE: comments in INI are ignored and lost when written.
	 * @throws IOException for any file IO errors
	 * @since 2.1
	 */
	public void write(PrintWriter f) throws IOException {
		for(int s = 0; s < sections.size(); s++) {
			Section section = (Section)sections.elementAt(s);
			f.println();
			f.println("[" + section.name + "]");
			for(int k = 0; k < section.size(); k++) {
				Key key = section.elementAt(k);
				f.println(key.name + "=" + key.value);
			}
		}
	}

	/**
	 * clears the contents of internal structures to can be reused.
	 * @since 2.1
	 */
	public void clear() {
		sections.removeAllElements();
	}

	/**
	 * sets the key to the specified value
	 * @param sectionName the name of the section for this key/value pair,
	 * will be created if necessary
	 * @param keyName the name of the key for this key/value pair, will be
	 * updated or added as necessary
	 * @param keyValue boolean value to set for this key/value pair
	 * @since 1.0
	 */
	public void setValue(String sectionName, String keyName, boolean keyValue) {
		String tf = "FALSE";
		if(keyValue) {
			tf = "TRUE";
		}
		setValue(sectionName, keyName, tf);
	}

	/**
	 * sets the key to the specified value
	 * @param sectionKeyPair the section key pair for this key/value pair,
	 * will be created if necessary
	 * @param keyValue boolean value to set for this key/value pair
	 * @since 1.0
	 */
	public void setValue(SectionKeyPair sectionKeyPair, boolean keyValue) {
		String tf = "FALSE";
		if(keyValue) {
			tf = "TRUE";
		}
		setValue(sectionKeyPair.sectionName, sectionKeyPair.keyName, tf);
	}

	/**
	 * sets the key to the specified value
	 * @param sectionName the name of the section for this key/value pair,
	 * will be created if necessary
	 * @param keyName the name of the key for this key/value pair, will be
	 * updated or added as necessary
	 * @param keyValue int value to set for this key/value pair
	 * @since 1.0
	 */
	public void setValue(String sectionName, String keyName, int keyValue) {
		setValue(sectionName, keyName, "" + keyValue);
	}

	/**
	 * sets the key to the specified value
	 * @param sectionKeyPair the section key pair for this key/value pair,
	 * will be created if necessary
	 * @param keyValue int value to set for this key/value pair
	 * @since 1.0
	 */
	public void setValue(SectionKeyPair sectionKeyPair, int keyValue) {
		setValue(sectionKeyPair.sectionName, sectionKeyPair.keyName, "" + keyValue);
	}

	/**
	 * sets the key to the specified value
	 * @param sectionName the name of the section for this key/value pair,
	 * will be created if necessary
	 * @param keyName the name of the key for this key/value pair, will be
	 * updated or added as necessary
	 * @param keyValue long value to set for this key/value pair
	 * @since 1.0
	 */
	public void setValue(String sectionName, String keyName, long keyValue) {
		setValue(sectionName, keyName, "" + keyValue);
	}

	/**
	 * sets the key to the specified value
	 * @param sectionKeyPair the section key pair for this key/value pair,
	 * will be created if necessary
	 * @param keyValue long value to set for this key/value pair
	 * @since 1.0
	 */
	public void setValue(SectionKeyPair sectionKeyPair, long keyValue) {
		setValue(sectionKeyPair.sectionName, sectionKeyPair.keyName, "" + keyValue);
	}

	/**
	 * sets the key to the specified value
	 * @param sectionName the name of the section for this key/value pair,
	 * will be created if necessary
	 * @param keyName the name of the key for this key/value pair, will be
	 * updated or added as necessary
	 * @param keyValue float value to set for this key/value pair
	 * @since 1.0
	 */
	public void setValue(String sectionName, String keyName, float keyValue) {
		setValue(sectionName, keyName, "" + keyValue);
	}

	/**
	 * sets the key to the specified value
	 * @param sectionKeyPair the section key pair for this key/value pair,
	 * will be created if necessary
	 * @param keyValue float value to set for this key/value pair
	 * @since 1.0
	 */
	public void setValue(SectionKeyPair sectionKeyPair, float keyValue) {
		setValue(sectionKeyPair.sectionName, sectionKeyPair.keyName, "" + keyValue);
	}

	/**
	 * sets the key to the specified value
	 * @param sectionName the name of the section for this key/value pair,
	 * will be created if necessary
	 * @param keyName the name of the key for this key/value pair, will be
	 * updated or added as necessary
	 * @param keyValue double value to set for this key/value pair
	 * @since 1.0
	 */
	public void setValue(String sectionName, String keyName, double keyValue) {
		setValue(sectionName, keyName, "" + keyValue);
	}

	/**
	 * sets the key to the specified value
	 * @param sectionKeyPair the section key pair for this key/value pair,
	 * will be created if necessary
	 * @param keyValue double value to set for this key/value pair
	 * @since 1.0
	 */
	public void setValue(SectionKeyPair sectionKeyPair, double keyValue) {
		setValue(sectionKeyPair.sectionName, sectionKeyPair.keyName, "" + keyValue);
	}

	/**
	 * sets the key to the specified value
	 * @param sectionKeyPair the section key pair for this key/value pair,
	 * will be created if necessary
	 * @param keyValue String value to set for this key/value pair
	 * @since 1.0
	 */
	public void setValue(SectionKeyPair sectionKeyPair, String keyValue) {
		setValue(sectionKeyPair.sectionName, sectionKeyPair.keyName, keyValue);
	}

	/**
	 * sets the key to the specified value
	 * @param sectionName the name of the section for this key/value pair,
	 * will be created if necessary
	 * @param keyName the name of the key for this key/value pair, will be
	 * updated or added as necessary
	 * @param keyValue String value to set for this key/value pair
	 * @since 1.0
	 */
	public void setValue(String sectionName, String keyName, String keyValue) {
		try {
			updateSection(getSection(sectionName), keyName, keyValue);

		}
		catch(SectionNotFoundException e) {
			currSection = new Section(sectionName);
			currSection.addKey(keyName, keyValue);
			sections.addElement(currSection);
		}
	}

	private Section getSection(String sectionName) throws SectionNotFoundException {
		if(currSection != null && currSection.name.equals(sectionName)) {
			return(currSection);
		}
		for(int s = 0; s < sections.size(); s++) {
			currSection = (Section)sections.elementAt(s);
			if(currSection.name.equals(sectionName)) {
				return(currSection);
			}
		}
		throw new SectionNotFoundException(sectionName);
	}

	private void updateSection(Section section, String keyName, String keyValue) {
		try {
			Key key = section.getKey(keyName);
			key.setValue(keyValue);
		}
		catch(KeyNotFoundException e) {
			section.addKey(keyName, keyValue);
		}
	}

	/**
	 * removes the section
	 * @param sectionName the name of the section
	 * @throws SectionNotFoundException if section cannot be found
	 * @since 1.0
	 */
	public void removeSection(String sectionName) throws SectionNotFoundException {
		if(currSection != null && currSection.name.equals(sectionName)) {
			sections.removeElement(currSection);
			currSection = null;
			return;
		}
		Section tempSection = null;
		for(int s = 0; s < sections.size(); s++) {
			tempSection = (Section)sections.elementAt(s);
			if(tempSection.name.equals(sectionName)) {
				sections.removeElementAt(s);
				return;
			}
		}
		throw new SectionNotFoundException(sectionName);
	}

	/**
	 * removes the key from the section
	 * @param sectionName the name of the section
	 * @param keyName the name of the key
	 * @throws SectionNotFoundException if section cannot be found
	 * @throws KeyNotFoundException if the key is not found
	 * @since 2.0
	 */
	public void removeKey(String sectionName, String keyName) throws SectionNotFoundException, KeyNotFoundException {
		getSection(sectionName).removeKey(keyName);
	}

	/**
	 * returns the boolean value of the key/value pair
	 * @param sectionName the name of the section for this key/value pair
	 * @param keyName the name of the key for this key/value pair
	 * @return the boolean value associated for this key/value pair
	 * @throws SectionNotFoundException if section cannot be found
	 * @throws KeyNotFoundException if key cannot be found
	 * @throws ValueNotFoundException if key has no value assigned
	 * @throws IllegalArgumentException if key has a illegal value
	 * @since 1.0
	 */
	public boolean getBooleanValue(String sectionName, String keyName) throws SectionNotFoundException, KeyNotFoundException, ValueNotFoundException {
		String value = getStringValue(sectionName, keyName);
		if(value.equalsIgnoreCase("TRUE")) {
			return(true);
		}
		else if(value.equalsIgnoreCase("FALSE")) {
			return(false);
		}
		else {
			throw new IllegalArgumentException(keyName);
		}
	}

	/**
	 * returns the boolean value of the key/value pair
	 * @param sectionKeyPair the section key pair for this key/value pair
	 * @return the boolean value associated for this key/value pair
	 * @throws SectionNotFoundException if section cannot be found
	 * @throws KeyNotFoundException if key cannot be found
	 * @throws ValueNotFoundException if key has no value assigned
	 * @throws IllegalArgumentException if key has a illegal value
	 * @since 1.0
	 */
	public boolean getBooleanValue(SectionKeyPair sectionKeyPair) throws SectionNotFoundException, KeyNotFoundException, ValueNotFoundException {
		return(getBooleanValue(sectionKeyPair.sectionName, sectionKeyPair.keyName));
	}

	/**
	 * returns the int value of the key/value pair
	 * @param sectionName the name of the section for this key/value pair
	 * @param keyName the name of the key for this key/value pair
	 * @return the int value associated for this key/value pair
	 * @throws SectionNotFoundException if section cannot be found
	 * @throws KeyNotFoundException if key cannot be found
	 * @throws ValueNotFoundException if key has no value assigned
	 * @throws NumberFormatException if the string does not contain a parsable
	 * integer. See <b>java.lang.Integer.parseLong</b> for more information.
	 * @since 1.0
	 */
	public int getIntValue(String sectionName, String keyName) throws SectionNotFoundException, KeyNotFoundException, ValueNotFoundException {
		return(new Integer(getStringValue(sectionName, keyName)).intValue());
	}

	/**
	 * returns the int value of the key/value pair
	 * @param sectionKeyPair the section key pair for this key/value pair
	 * @return the int value associated for this key/value pair
	 * @throws SectionNotFoundException if section cannot be found
	 * @throws KeyNotFoundException if key cannot be found
	 * @throws ValueNotFoundException if key has no value assigned
	 * @throws NumberFormatException if the string does not contain a parsable
	 * integer. See <b>java.lang.Integer.parseLong</b> for more information.
	 * @since 1.0
	 */
	public int getIntValue(SectionKeyPair sectionKeyPair) throws SectionNotFoundException, KeyNotFoundException, ValueNotFoundException {
		return(new Integer(getStringValue(sectionKeyPair.sectionName, sectionKeyPair.keyName)).intValue());
	}

	/**
	 * returns the long value of the key/value pair
	 * @param sectionName the name of the section for this key/value pair
	 * @param keyName the name of the key for this key/value pair
	 * @return the long value associated for this key/value pair
	 * @throws SectionNotFoundException if section cannot be found
	 * @throws KeyNotFoundException if key cannot be found
	 * @throws ValueNotFoundException if key has no value assigned
	 * @throws NumberFormatException if the string does not contain a parsable
	 * long. See <b>java.lang.Long.parseLong</b> for more information.
	 * @since 1.0
	 */
	public long getLongValue(String sectionName, String keyName) throws SectionNotFoundException, KeyNotFoundException, ValueNotFoundException {
		return(new Long(getStringValue(sectionName, keyName)).longValue());
	}

	/**
	 * returns the long value of the key/value pair
	 * @param sectionKeyPair the section key pair for this key/value pair
	 * @return the long value associated for this key/value pair
	 * @throws SectionNotFoundException if section cannot be found
	 * @throws KeyNotFoundException if key cannot be found
	 * @throws ValueNotFoundException if key has no value assigned
	 * @throws NumberFormatException if the string does not contain a parsable
	 * long. See <b>java.lang.Long.parseLong</b> for more information.
	 * @since 1.0
	 */
	public long getLongValue(SectionKeyPair sectionKeyPair) throws SectionNotFoundException, KeyNotFoundException, ValueNotFoundException {
		return(new Long(getStringValue(sectionKeyPair.sectionName, sectionKeyPair.keyName)).longValue());
	}

	/**
	 * returns the float value of the key/value pair
	 * @param sectionName the name of the section for this key/value pair
	 * @param keyName the name of the key for this key/value pair
	 * @return the float value associated for this key/value pair
	 * @throws SectionNotFoundException if section cannot be found
	 * @throws KeyNotFoundException if key cannot be found
	 * @throws ValueNotFoundException if key has no value assigned
	 * @throws NumberFormatException if the string does not contain a parsable
	 * long. See <b>java.lang.Float.parseLong</b> for more information.
	 * @since 1.0
	 */
	public float getFloatValue(String sectionName, String keyName) throws SectionNotFoundException, KeyNotFoundException, ValueNotFoundException {
		return(new Float(getStringValue(sectionName, keyName)).floatValue());
	}

	/**
	 * returns the float value of the key/value pair
	 * @param sectionKeyPair the section key pair for this key/value pair
	 * @return the float value associated for this key/value pair
	 * @throws SectionNotFoundException if section cannot be found
	 * @throws KeyNotFoundException if key cannot be found
	 * @throws ValueNotFoundException if key has no value assigned
	 * @throws NumberFormatException if the string does not contain a parsable
	 * long. See <b>java.lang.Float.parseLong</b> for more information.
	 * @since 1.0
	 */
	public float getFloatValue(SectionKeyPair sectionKeyPair) throws SectionNotFoundException, KeyNotFoundException, ValueNotFoundException {
		return(new Float(getStringValue(sectionKeyPair.sectionName, sectionKeyPair.keyName)).floatValue());
	}

	/**
	 * returns the double value of the key/value pair
	 * @param sectionName the name of the section for this key/value pair
	 * @param keyName the name of the key for this key/value pair
	 * @return the double value associated for this key/value pair
	 * @throws SectionNotFoundException if section cannot be found
	 * @throws KeyNotFoundException if key cannot be found
	 * @throws ValueNotFoundException if key has no value assigned
	 * @throws NumberFormatException if the string does not contain a parsable
	 * long. See <b>java.lang.Double.parseLong</b> for more information.
	 * @since 1.0
	 */
	public double getDoubleValue(String sectionName, String keyName) throws SectionNotFoundException, KeyNotFoundException, ValueNotFoundException {
		return(new Double(getStringValue(sectionName, keyName)).doubleValue());
	}

	/**
	 * returns the double value of the key/value pair
	 * @param sectionKeyPair the section key pair for this key/value pair
	 * @return the double value associated for this key/value pair
	 * @throws SectionNotFoundException if section cannot be found
	 * @throws KeyNotFoundException if key cannot be found
	 * @throws ValueNotFoundException if key has no value assigned
	 * @throws NumberFormatException if the string does not contain a parsable
	 * long. See <b>java.lang.Double.parseLong</b> for more information.
	 * @since 1.0
	 */
	public double getDoubleValue(SectionKeyPair sectionKeyPair) throws SectionNotFoundException, KeyNotFoundException, ValueNotFoundException {
		return(new Double(getStringValue(sectionKeyPair.sectionName, sectionKeyPair.keyName)).doubleValue());
	}

	/**
	 * returns the String value of the key/value pair
	 * @param sectionKeyPair the section key pair for this key/value pair
	 * @return the String value associated for this key/value pair
	 * @throws SectionNotFoundException if section cannot be found
	 * @throws KeyNotFoundException if key cannot be found
	 * @throws ValueNotFoundException if key has no value assigned
	 * @since 1.0
	 */
	public String getStringValue(SectionKeyPair sectionKeyPair) throws SectionNotFoundException, KeyNotFoundException, ValueNotFoundException {
		return(getStringValue(sectionKeyPair.sectionName, sectionKeyPair.keyName));
	}

	/**
	 * returns the String value of the key/value pair
	 * @param sectionName the name of the section for this key/value pair
	 * @param keyName the name of the key for this key/value pair
	 * @return the String value associated for this key/value pair
	 * @throws SectionNotFoundException if section cannot be found
	 * @throws KeyNotFoundException if key cannot be found
	 * @throws ValueNotFoundException if key has no value assigned
	 * @since 1.0
	 */
	public String getStringValue(String sectionName, String keyName) throws SectionNotFoundException, KeyNotFoundException, ValueNotFoundException {
		return(getValue(getSection(sectionName), keyName));
	}

	private String getValue(Section section, String keyName) throws KeyNotFoundException, ValueNotFoundException {
		Key key = section.getKey(keyName);
		if(key.value.equals("")) {
			throw new ValueNotFoundException(section + key.name);
		}
		return(key.value);
	}

	/**
	 * returns the list of section name
	 * @return the String array of section names
	 * @since 2.0
	 */
	public String [] getSections() {
		String [] names = null;
		if(sections.size() > 0) {
			names = new String [sections.size()];
			for(int s = 0; s < this.sections.size(); s++) {
				Section section = (Section)this.sections.elementAt(s);
				names[s] = section.name;
			}
		}
		return(names);
	}

	/**
	 * returns the list of keys for the section name
	 * @param sectionName the name of the section whos keys you wish returned
	 * @return the String array of key names
	 * @since 2.0
	 * @throws SectionNotFoundException if section cannot be found
	 */
	public String [] getKeys(String sectionName) throws SectionNotFoundException {
		return(getSection(sectionName).getKeys());
	}

	final private class Key {
		String name = "";
		String value = "";

		public Key(String name, String value) {
			this.name = name;
			this.value = value;
		}

		public String getName() {
			return(name);
		}

		public String getValue() {
			return(value);
		}

		public void setValue(String value) {
			this.value = value;
		}

		public String toString() {
			return(name + "=" + value);
		}

		public boolean equals(String name) {
			return(this.name.equals(name));
		}
	}

	final private class Section {
		String name = "";
		Vector keys = new Vector();
		Key currKey = null;

		public Section(String name) {
			this.name = name;
		}

		public String toString() {
			return("[" + name + "]");
		}

		public void addKey(String name, String value) {
			currKey = new Key(name, value);
			keys.addElement(currKey);
		}

		public Key getKey(String name) throws KeyNotFoundException {
			if(currKey != null && currKey.name.equals(name)) {
				return(currKey);
			}
			for(int k = 0; k < keys.size(); k++) {
				currKey = (Key)keys.elementAt(k);
				if(currKey.name.equals(name)) {
					return(currKey);
				}
			}
			throw new KeyNotFoundException(toString() + name);
		}

		public void removeKey(String name) throws KeyNotFoundException {
			if(currKey != null && currKey.name.equals(name)) {
				keys.removeElement(currKey);
				currKey = null;
				return;
			}
			for(int k = 0; k < keys.size(); k++) {
				Key key = (Key)keys.elementAt(k);
				if(key.name.equals(name)) {
					keys.removeElementAt(k);
					return;
				}
			}
			throw new KeyNotFoundException(toString() + name);
		}

		public Enumeration elements() {
			return(keys.elements());
		}

		public String [] getKeys() {
			String [] keys = null;
			if(this.keys.size() > 0) {
				keys = new String [this.keys.size()];
			}
			for(int k = 0; k < this.keys.size(); k++) {
				Key key = (Key)this.keys.elementAt(k);
				keys[k] = key.name;
			}
			return(keys);
		}

		public int size() {
			return(keys.size());
		}

		public Key elementAt(int k) {
			return(Key)keys.elementAt(k);
		}

		public boolean equals(String name) {
			return(this.name.equals(name));
		}
	}

	protected File iniFile = null;
	protected Vector sections = new Vector();
	protected Section currSection = null;
}

